@@ -2,7 +2,6 @@
 
 import os
 import warnings
-import re
 
 __all__ = ["getcaps","findmatch"]
 
@@ -14,11 +13,6 @@ def lineno_sort_key(entry):
     else:
         return 1, 0
 
-_find_unsafe = re.compile(r'[^\xa1-\U0010FFFF\w@+=:,./-]').search
-
-class UnsafeMailcapInput(Warning):
-    """Warning raised when refusing unsafe input"""
-
 
 # Part 1: top-level interface.
 
@@ -171,22 +165,15 @@ def findmatch(caps, MIMEtype, key='view', filename="/dev/null", plist=[]):
     entry to use.
 
     """
-    if _find_unsafe(filename):
-        msg = "Refusing to use mailcap with filename %r. Use a safe temporary filename." % (filename,)
-        warnings.warn(msg, UnsafeMailcapInput)
-        return None, None
     entries = lookup(caps, MIMEtype, key)
     # XXX This code should somehow check for the needsterminal flag.
     for e in entries:
         if 'test' in e:
             test = subst(e['test'], filename, plist)
-            if test is None:
-                continue
             if test and os.system(test) != 0:
                 continue
         command = subst(e[key], MIMEtype, filename, plist)
-        if command is not None:
-            return command, e
+        return command, e
     return None, None
 
 def lookup(caps, MIMEtype, key=None):
@@ -219,10 +206,6 @@ def subst(field, MIMEtype, filename, plist=[]):
             elif c == 's':
                 res = res + filename
             elif c == 't':
-                if _find_unsafe(MIMEtype):
-                    msg = "Refusing to substitute MIME type %r into a shell command." % (MIMEtype,)
-                    warnings.warn(msg, UnsafeMailcapInput)
-                    return None
                 res = res + MIMEtype
             elif c == '{':
                 start = i
@@ -230,12 +213,7 @@ def subst(field, MIMEtype, filename, plist=[]):
                     i = i+1
                 name = field[start:i]
                 i = i+1
-                param = findparam(name, plist)
-                if _find_unsafe(param):
-                    msg = "Refusing to substitute parameter %r (%s) into a shell command" % (param, name)
-                    warnings.warn(msg, UnsafeMailcapInput)
-                    return None
-                res = res + param
+                res = res + findparam(name, plist)
             # XXX To do:
             # %n == number of parts if type is multipart/*
             # %F == list of alternating type and filename for parts
